from .metadata import MetadataMapping
from .risearch import ResourceIndexSearch
from .finder import FileOrganizer
from .curate import FileCurator
from .validate import ValidateMigration
from .fedora import FedoraObject
from.checksum import HashSheet
from .controller import InterfaceController
from .combine import ImportRefactor
from .template import ImportTemplate
from .restrict import Restrictions, RestrictionsSheet

__all__ = [
    "FedoraObject",
    "FileCurator",
    "FileOrganizer",
    "HashSheet",
    "ImportRefactor",
    "ImportTemplate",
    "InterfaceController",
    "MetadataMapping",
    "ResourceIndexSearch",
    "Restrictions",
    "RestrictionsSheet",
    "ValidateMigration",
]



import hashlib
from csv import DictWriter, DictReader
import os
import requests
from tqdm import tqdm


class HashSheet:
    def __init__(self, path, output):
        self.path = path
        self.output = output
        self.all_files = self.walk_sheets(path)

    @staticmethod
    def walk_sheets(path):
        """Walk through a directory and return a list of all files.

        Args:
            path (str): The path to the directory to walk through.

        Returns:
            list: A list of all files in the directory.

        Examples:
            >>> hs = HashSheet("tests/fixtures/bad_imports", "example.csv")
            >>> hs.walk_sheets("tests/fixtures/bad_imports")
            ['https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:248_MODS.xml', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:309_MODS.xml', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:118_OBJ.jp2', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:99_MODS.xml', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:374_MODS.xml', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:465_OBJ.jp2', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:356_MODS.xml', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:254_OBJ.jp2', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:451_OBJ.jp2', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:104_MODS.xml', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:419_MODS.xml', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:318_OBJ.jp2', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:463_MODS.xml', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:350_MODS.xml', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:108_OBJ.jp2', 'https://digital.lib.utk.edu/migration/mpaekefauver/mpaekefauver:420_MODS.xml']

        """
        all_files = []
        for path, directories, files in os.walk(path):
            for filename in files:
                with open(f"{path}/{filename}", "r") as f:
                    reader = DictReader(f)
                    for row in reader:
                        all_files.append(row["remote_files"])
        return all_files

    def checksum(self):
        """Calculate the sha1 checksum of all files listed in csvs in a directory.

        Returns:
            list: A list of dictionaries with the url and checksum of each file.

        Examples:
            No example to keep tests running quickly.

        """
        files_with_checksums = []
        for file in tqdm(self.all_files):
            hash = self.checksum_file(file)
            files_with_checksums.append(hash)
        return files_with_checksums

    @staticmethod
    def checksum_file(file):
        """Calculate the sha1 checksum of a file.

        Args:
            file (str): The path to the file to checksum.

        Returns:
            dict: A dictionary with the url and checksum of the file.

        Examples:
            >>> hs = HashSheet("tests/fixtures/bad_imports", "example.csv")
            >>> hs.checksum_file("https://raw.githubusercontent.com/utkdigitalinitiatives/utk-exodus/main/tests/fixtures/colloquy_202.xml")
            {'url': 'https://raw.githubusercontent.com/utkdigitalinitiatives/utk-exodus/main/tests/fixtures/colloquy_202.xml', 'checksum': '081a51fae0200f266d2933756d48441c4ea77b1e'}

        """
        response = requests.get(file, stream=True)
        response.raise_for_status()
        sha1 = hashlib.sha1()
        for chunk in response.iter_content(chunk_size=8192):
            if chunk:
                sha1.update(chunk)
        return {"url": file, "checksum": sha1.hexdigest()}

    def write(self):
        with open(self.output, "w") as csvfile:
            writer = DictWriter(csvfile, fieldnames=["url", "checksum"])
            writer.writeheader()
            writer.writerows(self.checksum())
        return


if __name__ == "__main__":
    path = "tests/fixtures/bad_imports"
    output = "delete/sample_checksums.csv"
    checksum = HashSheet(path, output)
    checksum.write()



import yaml
from csv import DictWriter


class ImportTemplate:
    def __init__(self, profile, model):
        self.profile = profile
        self.model = model
        self.loaded_m3 = yaml.safe_load(open(profile))
        self.headers = self.__get_all_properties()

    def __get_base_properties(self):
        return [
            "source_identifier",
            "model",
            "sequence",
            "remote_files",
            "parents",
            "visibility",
        ]

    def __get_all_properties(self):
        properties = self.__get_base_properties()
        for k, v in self.loaded_m3["properties"].items():
            if self.model in v["available_on"]["class"]:
                properties.append(k)
        return properties

    def add_cardinality(self):
        row = {}
        for k, v in self.loaded_m3["properties"].items():
            if self.model in v["available_on"]["class"]:
                row[k] = (
                    f"Minimum: {v['cardinality'].get('minimum', 0)} / Maximum: {v['cardinality'].get('maximum', 'n')}"
                )
                row[k] = (
                    f"{v['cardinality'].get('minimum', 0)}, {v['cardinality'].get('maximum', 'n')}"
                )
        return row

    def add_range(self):
        row = {}
        for k, v in self.loaded_m3["properties"].items():
            if self.model in v["available_on"]["class"]:
                row[k] = v.get("range").split("#")[-1]
        return row

    def add_guidelines(self):
        row = {}
        for k, v in self.loaded_m3["properties"].items():
            if self.model in v["available_on"]["class"]:
                if "usage_guidelines" in v:
                    row[k] = v.get("usage_guidelines").get("default", "")
                elif "definition" in v:
                    row[k] = v.get("definition").get("default", "")
                else:
                    row[k] = ""
        return row

    def write(self, filename):
        with open(filename, "w") as f:
            writer = DictWriter(f, fieldnames=self.headers)
            writer.writeheader()
            writer.writerow(self.add_cardinality())
            writer.writerow(self.add_range())
            writer.writerow(self.add_guidelines())


if __name__ == "__main__":
    it = ImportTemplate("tmp/m3.yml", "Image")
    it.write("full_template.csv")



import xml.etree.ElementTree as ET
import csv
from tqdm import tqdm
import os


class Restrictions:
    def __init__(self, policy):
        self.policy = policy
        self.ns = {"xacml": "urn:oasis:names:tc:xacml:1.0:policy"}
        self.root = self.__parse(policy)

    @staticmethod
    def __parse(policy):
        tree = ET.parse(policy)
        return tree.getroot()

    def find_rules(self):
        rules = []
        for rule in self.root.findall(".//xacml:Rule", self.ns):
            rule_id = rule.get("RuleId")
            effect = rule.get("Effect")
            rules.append((rule_id, effect))
        return rules

    def find_restricted_datastreams(self):
        restricted_datastreams = []
        for rule in self.root.findall(
            './/xacml:Rule[@RuleId="deny-dsid-mime"]', self.ns
        ):
            for resource in rule.findall(".//xacml:Resource", self.ns):
                dsid = resource.find(".//xacml:AttributeValue", self.ns).text
                restricted_datastreams.append(dsid)
        return restricted_datastreams

    def find_objects_only_accessible_by_certain_users(self):
        users_with_access = []
        for rule in self.root.findall(
            './/xacml:Rule[@RuleId="deny-access-functions"]', self.ns
        ):
            for condition in rule.findall(
                './/xacml:Condition[@FunctionId="urn:oasis:names:tc:xacml:1.0:function:not"]',
                self.ns,
            ):
                for apply in condition.findall(
                    './/xacml:Apply[@FunctionId="urn:oasis:names:tc:xacml:1.0:function:string-at-least-one-member-of"]',
                    self.ns,
                ):
                    users_with_access.append(
                        apply.find(".//xacml:AttributeValue", self.ns).text
                    )
        return users_with_access

    def determine_if_work_restricted(self):
        if len(self.find_objects_only_accessible_by_certain_users()) > 0:
            return True
        return False

    def get(self):
        return {
            "work_restricted": self.determine_if_work_restricted(),
            "restricted_datastreams": self.find_restricted_datastreams(),
        }


class RestrictionsSheet:
    def __init__(self, original_sheet, policies_location):
        self.original_sheet = original_sheet
        self.policies_location = policies_location
        self.original_as_dict = self.__read(original_sheet)
        self.headers = self.__get_headers()
        self.rows_with_visibility = self.add_visibility()

    @staticmethod
    def __read(csv_file):
        csv_content = []
        with open(csv_file, "r") as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                csv_content.append(row)
        return csv_content

    def __get_headers(self):
        original_headers = [k for k, v in self.original_as_dict[0].items()]
        original_headers.append("visibility")
        return original_headers

    def add_visibility(self):
        new_csv_content = []
        for row in tqdm(self.original_as_dict):
            pid = row["source_identifier"].split("_")[0]
            datastream = None
            visibility = "open"
            if len(row["source_identifier"].split("_")) > 1:
                datastream = row["source_identifier"].split("_")[1]
            if os.path.exists(f"{self.policies_location}/{pid}_POLICY.xml"):
                restrictions = Restrictions(
                    f"{self.policies_location}/{pid}_POLICY.xml"
                ).get()
                if restrictions["work_restricted"]:
                    visibility = "restricted"
                elif datastream in restrictions["restricted_datastreams"]:
                    visibility = "restricted"
            current_data = {}
            for k, v in row.items():
                current_data[k] = v
            current_data["visibility"] = visibility
            new_csv_content.append(current_data)
        return new_csv_content

    def write_csv(self, filename):
        with open(filename, "w", newline="") as bulkrax_sheet:
            writer = csv.DictWriter(bulkrax_sheet, fieldnames=self.headers)
            writer.writeheader()
            for data in self.rows_with_visibility:
                writer.writerow(data)
        return





import requests
from urllib.parse import quote


class ResourceIndexSearch:
    def __init__(
        self,
        language="sparql",
        riformat="CSV",
        ri_endpoint="https://porter.lib.utk.edu/fedora/risearch",
    ):
        self.risearch_endpoint = ri_endpoint
        self.valid_languages = ("itql", "sparql")
        self.valid_formats = ("CSV", "Simple", "Sparql", "TSV", "JSON")
        self.language = self.validate_language(language)
        self.format = self.validate_format(riformat)
        self.base_url = (
            f"{self.risearch_endpoint}?type=tuples"
            f"&lang={self.language}&format={self.format}"
        )

    def validate_language(self, language):
        if language in self.valid_languages:
            return language
        else:
            raise Exception(
                f"Supplied language is not valid: {language}. Must be one of {self.valid_languages}."
            )

    def validate_format(self, user_format):
        if user_format in self.valid_formats:
            return user_format
        else:
            raise Exception(
                f"Supplied format is not valid: {user_format}. Must be one of {self.valid_formats}."
            )

    @staticmethod
    def __clean_csv_results(split_results, uri_prefix):
        results = []
        for result in split_results:
            if result.startswith(uri_prefix):
                new_result = result.split(",")
                results.append(
                    (new_result[0].replace(uri_prefix, ""), int(new_result[1]))
                )
        return sorted(results, key=lambda x: x[1])

    def get_files(self, pid):
        if self.language != "sparql":
            raise Exception(
                f"You must use sparql as the language for this method.  You used {self.language}."
            )
        sparql_query = quote(
            f"SELECT $files FROM <#ri> WHERE {{ <info:fedora/{pid}> "
            f"<info:fedora/fedora-system:def/view#disseminates> $files . }}"
        )
        results = (
            requests.get(f"{self.base_url}&query={sparql_query}")
            .content.decode("utf-8")
            .split("\n")
        )
        return [
            result.split("/")[-1] for result in results if result.startswith("info")
        ]

    def __request_pids(self, request):
        results = (
            requests.get(f"{self.base_url}&query={request}")
            .content.decode("utf-8")
            .split("\n")
        )
        return [
            result.split("/")[-1] for result in results if result.startswith("info")
        ]

    def __request_json(self, request):
        return requests.get(f"{self.base_url}&query={request}").json()

    def get_images_no_parts(self, collection):
        """@Todo: Fix long queries."""
        members_of_collection_query = quote(
            f"SELECT ?pid FROM <#ri> WHERE "
            f"{{ ?pid <info:fedora/fedora-system:def/model#hasModel> <info:fedora/islandora:sp_large_image_cmodel> ; "
            f"<info:fedora/fedora-system:def/relations-external#isMemberOfCollection> <info:fedora/{collection}> . }}"
        )
        members_of_collection_parts_query = quote(
            f"SELECT ?pid FROM <#ri> WHERE "
            f"{{ ?pid <info:fedora/fedora-system:def/model#hasModel> <info:fedora/islandora:sp_large_image_cmodel> ; "
            f"<info:fedora/fedora-system:def/relations-external#isMemberOfCollection> <info:fedora/{collection}> ; "
            f"<info:fedora/fedora-system:def/relations-external#isConstituentOf> ?unknown. }}"
        )
        members_of_collection = self.__request_pids(members_of_collection_query)
        members_of_collection_parts = self.__request_pids(
            members_of_collection_parts_query
        )
        members_of_collection_no_parts = [
            pid
            for pid in members_of_collection
            if pid not in members_of_collection_parts
        ]
        return members_of_collection_no_parts

    def get_parent_collections(self, pid):
        query = quote(
            f"""SELECT ?parent FROM <#ri> WHERE {{<info:fedora/{pid}> <info:fedora/fedora-system:def/relations-external#isMemberOfCollection> ?parent .}}"""
        )
        collections = self.__request_pids(query)
        return collections

    def get_members_types_and_collections(self, pid):
        query = quote(
            f"""SELECT ?pid ?work_type ?collection FROM <#ri> WHERE {{?pid <info:fedora/fedora-system:def/relations-external#isMemberOfCollection> <info:fedora/{pid}> ;<info:fedora/fedora-system:def/model#hasModel> ?work_type ;<info:fedora/fedora-system:def/relations-external#isMemberOfCollection> ?collection .}}"""
        )
        results = self.__request_json(query)
        return results

    def get_islandora_work_type(self, pid):
        query = quote(
            f"""SELECT ?work_type FROM <#ri> WHERE {{<info:fedora/{pid}> <info:fedora/fedora-system:def/model#hasModel> ?work_type .}}"""
        )
        results = requests.get(f"{self.base_url}&query={query}").content.decode("utf-8")
        return [
            result.strip()
            for result in results.split("\n")
            if "info:fedora/fedora-system:FedoraObject-3.0" not in result
        ][1]

    def count_books_and_pages_in_collection(self, collection):
        pages = []
        query = quote(
            f"""SELECT ?pid FROM <#ri> WHERE {{ ?pid <info:fedora/fedora-system:def/relations-external#isMemberOfCollection> <info:fedora/{collection}>; <info:fedora/fedora-system:def/model#hasModel> <info:fedora/islandora:bookCModel> .}}"""
        )
        results = requests.get(f"{self.base_url}&query={query}").json()
        books = [result["pid"] for result in results["results"]]
        for book in books:
            query = quote(
                f"""SELECT ?page FROM <#ri> WHERE {{?page <http://islandora.ca/ontology/relsext#isPageOf> <{book}>.}}"""
            )
            page_results = requests.get(f"{self.base_url}&query={query}").json()
            for page in page_results["results"]:
                pages.append(page)
        return len(books), len(pages)

    def find_pages_in_book(self, book):
        query = quote(
            f"SELECT ?pid ?page WHERE {{ "
            f"?pid <info:fedora/fedora-system:def/model#hasModel> "
            f"<info:fedora/islandora:pageCModel> ;"
            f"<info:fedora/fedora-system:def/relations-external#isMemberOf> "
            f"<info:fedora/{book}> ; "
            f"<http://islandora.ca/ontology/relsext#isPageNumber> ?page . }}"
        )
        page_results = requests.get(f"{self.base_url}&query={query}").content
        return self.clean_pages(page_results)

    @staticmethod
    def clean_pages(results):
        all_pages = []
        cleaned = results.decode("utf-8").split("\n")
        for item in cleaned:
            if item != '"pid","page"' and item != "":
                all_pages.append(
                    {"pid": item.split(",")[0], "page": item.split(",")[1]}
                )
        return all_pages

    @staticmethod
    def __lookup_work_type(work_type):
        work_types = {
            "book": "info:fedora/islandora:bookCModel",
            "image": "info:fedora/islandora:sp_basic_image",
            "large_image": "info:fedora/islandora:sp_large_image_cmodel",
            "compound": "info:fedora/islandora:compoundCModel",
            "audio": "info:fedora/islandora:sp-audioCModel",
            "video": "info:fedora/islandora:sp_videoCModel",
            "pdf": "info:fedora/islandora:sp_pdf",
            "page": "info:fedora/islandora:pageCModel",
        }
        return work_types.get(work_type, "unknown")

    def get_works_based_on_type_and_collection(self, work_type, collection):
        iri = self.__lookup_work_type(work_type).strip()
        query = quote(
            f"PREFIX rels-ext: <info:fedora/fedora-system:def/relations-external#>"
            f"PREFIX model: <info:fedora/fedora-system:def/model#>"
            f"SELECT ?pid WHERE {{ ?pid rels-ext:isMemberOfCollection <info:fedora/{collection}> ;"
            f"model:hasModel <{iri}> ."
            f"}}"
        )
        results = requests.get(f"{self.base_url}&query={query}").content.decode("utf-8")
        return [result for result in results.split("\n") if result != "" and result != '"pid"']

    def get_policies_for_pages_in_book(self, book):
        query = quote(
            f"PREFIX system: <info:fedora/fedora-system:def/view#>"
            f"PREFIX rels-ext: <info:fedora/fedora-system:def/relations-external#>"
            f"PREFIX model: <info:fedora/fedora-system:def/model#>"
            f"SELECT ?pid WHERE {{"
            f"?pid system:disseminates ?o ;"
            f"rels-ext:isMemberOf <{book}> ;"
            f"model:hasModel <info:fedora/islandora:pageCModel> ."
            f"FILTER(REGEX(STR(?o), 'POLICY')).}}"
        )
        results = requests.get(f"{self.base_url}&query={query}").content.decode("utf-8")
        return [result for result in results.split("\n") if result != "" and result != '"pid"']

    def get_policies_based_on_type_and_collection(self, work_type, collection):
        iri = self.__lookup_work_type(work_type).strip()
        query = quote(
            f"PREFIX system: <info:fedora/fedora-system:def/view#>"
            f"PREFIX rels-ext: <info:fedora/fedora-system:def/relations-external#>"
            f"PREFIX model: <info:fedora/fedora-system:def/model#>"
            f"SELECT ?pid WHERE {{"
            f"?pid system:disseminates ?o ;"
            f"rels-ext:isMemberOfCollection <info:fedora/{collection}> ;"
            f"model:hasModel <{iri}> ."
            f"FILTER(REGEX(STR(?o), 'POLICY')).}}"
        )
        results = requests.get(f"{self.base_url}&query={query}").content.decode("utf-8")
        if work_type != "book":
            return [result for result in results.split("\n") if result != "" and result != '"pid"']
        else:
            all_policies_from_book = []
            books = [result for result in results.split("\n") if result != "" and result != '"pid"']
            for book in books:
                all_policies_from_book.append(book)
                all_policies_from_book.extend(self.get_policies_for_pages_in_book(book))
            return all_policies_from_book

    def get_parent_book(self, pid):
        query = quote(
            f"SELECT ?book FROM <#ri> WHERE {{"
            f"<info:fedora/{pid}> <http://islandora.ca/ontology/relsext#isPageOf> ?book ."
            f"}}"
        )
        results = requests.get(f"{self.base_url}&query={query}").content.decode("utf-8")
        return results.split("\n")[1]

    def get_page_number(self, pid):
        query = quote(
            f"SELECT ?page FROM <#ri> WHERE {{"
            f"<info:fedora/{pid}> <http://islandora.ca/ontology/relsext#isPageNumber> ?page ."
            f"}}"
        )
        results = requests.get(f"{self.base_url}&query={query}").content.decode("utf-8")
        return results.split("\n")[1]


if __name__ == "__main__":
    risearch = ResourceIndexSearch()
    x = risearch.get_policies_based_on_type_and_collection("book", "collections:galston")
    print(x)



import yaml
import csv


class ValidateMigration:
    def __init__(self, profile, migration_sheet):
        self.csv = migration_sheet
        self.loaded_csv = self.__read_csv()
        self.profile = profile
        self.loaded_m3 = yaml.safe_load(open(profile))
        self.all_exceptions = []

    def __read_csv(self):
        csv_content = []
        with open(self.csv, "r") as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                csv_content.append(row)
        return csv_content

    def validate_model(self, row):
        models = [k for k, v in self.loaded_m3["classes"].items()]
        if row["model"] != "FileSet" and row["model"] != "Collection":
            if row["model"] not in models:
                self.all_exceptions.append(
                    f'{row["source_identifier"]} has invalid model {row["model"]}.'
                )
        return

    def validate_license(self, row):
        if "license" in row and row["license"] != "":
            if "http://" not in row["license"]:
                self.all_exceptions.append(
                    f'{row["source_identifier"]} has invalid license: {row["license"]}.'
                )
            if "/rdf" in row["license"]:
                self.all_exceptions.append(
                    f'{row["source_identifier"]} has invalid license: {row["license"]}.'
                )

    def validate_values(self, row):
        system_fields = (
            "source_identifier",
            "model",
            "remote_files",
            "parents",
            "visibility",
        )
        for k, v in row.items():
            if (
                k not in system_fields
                and row["model"] != "FileSet"
                and row["model"] != "Collection"
            ):
                review_property = self.check_property(k)
                if review_property:
                    available_on = self.check_available_on(row, k, v)
                    if available_on:
                        self.check_cardinality(row, k, v)
                        self.check_range(k, v, row)
        return

    def check_available_on(self, row, key, value):
        if (
            row["model"]
            not in self.loaded_m3["properties"][key]["available_on"]["class"]
            and value != ""
        ):
            self.all_exceptions.append(
                f"{key} is not available on {row['model']} for {row['source_identifier']}."
            )
            return False
        elif (
            row["model"]
            not in self.loaded_m3["properties"][key]["available_on"]["class"]
        ):
            return False
        else:
            return True

    def check_property(self, key):
        if key not in self.loaded_m3["properties"]:
            self.all_exceptions.append(f"{key} is not listed in the m3 profile.")
            return False
        else:
            return True

    def check_cardinality(self, row, key, value):
        maximum = (
            self.loaded_m3["properties"][key]["cardinality"]["maximum"]
            if "maximum" in self.loaded_m3["properties"][key]["cardinality"]
            else 1000
        )
        minimum = (
            self.loaded_m3["properties"][key]["cardinality"]["minimum"]
            if "minimum" in self.loaded_m3["properties"][key]["cardinality"]
            else 0
        )
        all_values = [thing for thing in value.split(" | ") if thing != ""]
        if len(all_values) > maximum:
            self.all_exceptions.append(
                f'{key} has {len(all_values)} values but maximum is {maximum} on {row["source_identifier"]}.'
            )
        if len(all_values) < minimum:
            self.all_exceptions.append(
                f'{key} has {len(all_values)} values but minimum is {minimum} on {row["model"]} on {row["source_identifier"]}.'
            )
        return

    def check_range(self, key, value, row):
        all_values = [thing for thing in value.split(" | ") if thing != ""]
        property_range = self.loaded_m3["properties"][key]["range"]
        for value in all_values:
            if property_range == "http://www.w3.org/2001/XMLSchema#anyURI":
                if not (value.startswith("http")):
                    self.all_exceptions.append(
                        f'{value} is not a URI for {row["source_identifier"]}.'
                    )
            else:
                if value.startswith("http:"):
                    self.all_exceptions.append(
                        f'{value} may be a URI for {row["source_identifier"]}.'
                    )
        return

    def check_required_fields_are_present(self, row):
        for k, v in self.loaded_m3["properties"].items():
            if "minimum" in v["cardinality"]:
                if (
                    v["cardinality"]["minimum"] > 0
                    and row["model"] in v["available_on"]["class"]
                ):
                    keys = [key for key, value in row.items()]
                    if k not in keys:
                        self.all_exceptions.append(
                            f'{row["source_identifier"]} has no {k} but {k} required on {row["model"]}'
                        )

    def iterate(self):
        for row in self.loaded_csv:
            self.validate_model(row)
            self.validate_values(row)
            self.validate_license(row)
            self.check_required_fields_are_present(row)
        separator = "\n"
        if len(self.all_exceptions) > 0:
            raise Exception(
                f"\tMigration spreadsheet has at least {len(self.all_exceptions)} problems: {separator.join(self.all_exceptions)}"
            )
        else:
            print("\tSheet passes all tests.")


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description="Choose sheet to validate..")
    parser.add_argument(
        "-s", "--sheet", dest="sheet", help="Specify csv to test.", required=True
    )
    parser.add_argument(
        "-p", "--profile", dest="profile", help="Specify a profile.", default="current"
    )
    args = parser.parse_args()
    x = ValidateMigration(profile=args.profile, migration_sheet=args.sheet)
    x.iterate()



import csv
from tqdm import tqdm
from utk_exodus.risearch import ResourceIndexSearch


class FileOrganizer:
    def __init__(
            self,
            csv,
            what_to_add=['filesets', 'attachments'],
            remote='https://digital.lib.utk.edu/collections/islandora/object/'
    ):
        self.original_csv = csv
        self.remote = remote
        self.original_as_dict = self.__read()
        self.headers = self.__get_headers()
        self.new_csv_with_files = self.__add_files(what_to_add)

    def __read(self):
        csv_content = []
        with open(self.original_csv, 'r') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                csv_content.append(row)
        return csv_content

    def __get_headers(self):
        original_headers = [k for k, v in self.original_as_dict[0].items()]
        original_headers.append('rdf_type')
        original_headers.append('file_language')
        return original_headers

    def __add_a_file(self, filename, row, preserve_and_obj=False, parent=""):
        default_headings = ('source_identifier', 'sequence', 'model', 'remote_files', 'title', 'abstract', 'parents', 'rdf_type')
        initial_data = {
            'source_identifier': f"{row['source_identifier'].replace('.xml', '')}_{filename}_fileset",
            'model': "FileSet",
            'sequence': row['sequence'],
            'remote_files': f"{self.remote}{row['source_identifier'].replace('_MODS.xml', '').replace('_', ':').replace('.xml', '')}/datastream/{filename}",
            'title': self.__get_filename_title(filename, preserve_and_obj, row),
            'abstract': f"{filename} for {row['source_identifier']}",
            'parents': f"{row['source_identifier'].replace('.xml', '')}_{filename}",
            'rdf_type': RDFTypeGenerator(row['model']).find_file_types(filename, preserve_and_obj),
            'file_language': ''
        }
        if parent != "":
            initial_data['parents'] = parent
        if preserve_and_obj is True and filename == "OBJ" and row['model'] == "Pdf":
            initial_data['parents'] = f"{row['source_identifier'].replace('.xml', '')}"
        if row['model'] == "Pdf" and filename in ("OBJ", "PDFA"):
            initial_data['remote_files'] = f"https://digital.lib.utk.edu/collections/islandora/object/{row['source_identifier'].replace('_MODS.xml', '').replace('_', ':').replace('.xml', '')}/datastream/{filename}/view.pdf"
        for k, v in row.items():
            if k not in default_headings:
                initial_data[k] = ''
        if row['model'] == "Audio" or row['model'] == "Video":
            if filename == "TRANSCRIPT":
                initial_data['file_language'] = 'en'
                initial_data['title'] = 'Captions in English'
            elif filename == "TRANSCRIPT-ES":
                initial_data['file_language'] = 'es'
                initial_data['title'] = 'Subtítulos en español'
        return initial_data

    def __add_an_attachment(self, filename, row, preserve_and_obj=False, parent=""):
        default_headings = ('source_identifier', 'sequence', 'model', 'remote_files', 'title', 'abstract', 'parents', 'rdf_type')
        initial_data = {
            'source_identifier': f"{row['source_identifier'].replace('.xml', '')}_{filename}",
            'model': "Attachment",
            'sequence': row['sequence'],
            'remote_files': "",
            'title': self.__get_filename_title(filename, preserve_and_obj, row),
            'abstract': f"{filename} for {row['source_identifier']}",
            'parents': row['source_identifier'].replace('.xml', ''),
            'rdf_type': RDFTypeGenerator(row['model']).find_file_types(filename, preserve_and_obj),
            'file_language': ''
        }
        if parent != "":
            initial_data['parents'] = parent
        for k, v in row.items():
            if k not in default_headings:
                initial_data[k] = ''
        if row['model'] == "Audio" or row['model'] == "Video":
            if filename == "TRANSCRIPT":
                initial_data['file_language'] = 'en'
                initial_data['title'] = 'Captions in English'
            elif filename == "TRANSCRIPT-ES":
                initial_data['file_language'] = 'es'
                initial_data['title'] = 'Subtítulos en español'
        return initial_data

    @staticmethod
    def __get_filename_title(dsid, preserve_and_obj, row):
        if preserve_and_obj and row['model'] == "Image":
            identifier = row['local_identifier'].split(' | ')[0]
            if dsid == "OBJ":
                return f"{identifier}_i"
            elif dsid == "PRESERVE":
                return f"{identifier}_p"
            else:
                return dsid
        elif preserve_and_obj and row['model'] == "Audio":
            identifier = row['local_identifier'].split(' | ')[0]
            if dsid == "OBJ":
                return f"{identifier}_p"
            elif dsid == "PROXY_MP3":
                return f"{identifier}_i"
            else:
                return dsid
        else:
            return dsid

    @staticmethod
    def __get_rdf_types_for_file(dsid, preserve_and_obj):
        if dsid == "OBJ" and preserve_and_obj is False:
            return "http://pcdm.org/use#PreservationFile | http://pcdm.org/use#IntermediateFile"
        elif dsid == "OBJ":
            return "http://pcdm.org/use#IntermediateFile"
        elif dsid == "PRESERVE":
            return "http://pcdm.org/use#PreservationFile"
        elif dsid == "MODS":
            return "http://pcdm.org/file-format-types#Markup"
        elif dsid == "POLICY":
            return "http://pcdm.org/file-format-types#StructuredText"
        elif dsid == "OCR":
            return "http://pcdm.org/use#ExtractedText"
        elif dsid == "HOCR":
            return "http://pcdm.org/file-format-types#HTML"
        else:
            return "http://pcdm.org/use#OriginalFile"

    def __add_files(self, what_to_add=['filesets', 'attachments']):
        new_csv_content = []
        for row in tqdm(self.original_as_dict):
            if row['model'] != "Page":
                new_csv_content.append(row)
            pid = row['source_identifier'].replace('_MODS.xml', '').replace('_', ":")
            all_files = FileSetFinder(pid).files
            if row['model'] == "Image":
                for dsid in all_files:
                    if 'PRESERVE' in all_files and 'OBJ' in all_files:
                        if 'attachments' in what_to_add:
                            new_csv_content.append(self.__add_an_attachment(dsid, row, True))
                        if 'filesets' in what_to_add:
                            new_csv_content.append(self.__add_a_file(dsid, row, True))
                    else:
                        if 'attachments' in what_to_add:
                            new_csv_content.append(self.__add_an_attachment(dsid, row))
                        if 'filesets' in what_to_add:
                            new_csv_content.append(self.__add_a_file(dsid, row))
            elif row['model'] == "Audio":
                for dsid in all_files:
                    if 'OBJ' in all_files and 'PROXY_MP3' in all_files:
                        new_csv_content.append(self.__add_an_attachment(dsid, row, True))
                        new_csv_content.append(self.__add_a_file(dsid, row, True))
                    else:
                        new_csv_content.append(self.__add_an_attachment(dsid, row))
                        new_csv_content.append(self.__add_a_file(dsid, row))
            elif row['model'] == "Video":
                for dsid in all_files:
                    if 'OBJ' in all_files and 'MP4' in all_files:
                        new_csv_content.append(self.__add_an_attachment(dsid, row, True))
                        new_csv_content.append(self.__add_a_file(dsid, row, True))
                    else:
                        new_csv_content.append(self.__add_an_attachment(dsid, row))
                        new_csv_content.append(self.__add_a_file(dsid, row))
            elif row['model'] == "Pdf":
                for dsid in all_files:
                    if 'OBJ' in all_files and 'PDFA' in all_files:
                        if dsid == "OBJ":
                            new_csv_content.append(self.__add_a_file(dsid, row, True))
                        else:
                            new_csv_content.append(self.__add_an_attachment(dsid, row, True))
                            new_csv_content.append(self.__add_a_file(dsid, row, True))
                    else:
                        new_csv_content.append(self.__add_an_attachment(dsid, row))
                        new_csv_content.append(self.__add_a_file(dsid, row))
            elif row['model'] == "Collection":
                pass
            elif row['model'] == "Book":
                for dsid in all_files:
                    if 'PRESERVE' in all_files and 'OBJ' in all_files:
                        if 'attachments' in what_to_add:
                            new_csv_content.append(self.__add_an_attachment(dsid, row, True))
                        if 'filesets' in what_to_add:
                            new_csv_content.append(self.__add_a_file(dsid, row, True))
                    else:
                        if 'attachments' in what_to_add:
                            new_csv_content.append(self.__add_an_attachment(dsid, row))
                        if 'filesets' in what_to_add:
                            new_csv_content.append(self.__add_a_file(dsid, row))
            elif row['model'] == "Page":
                dsids_to_remove = ('MODS', 'RELS-INT', 'PDF')
                for dsid in dsids_to_remove:
                    if dsid in all_files:
                        all_files.remove(dsid)
                for dsid in all_files:
                    if 'PRESERVE' in all_files and 'OBJ' in all_files:
                        if 'attachments' in what_to_add:
                            new_csv_content.append(self.__add_an_attachment(dsid, row, True, parent=row['parents']))
                        if 'filesets' in what_to_add:
                            new_csv_content.append(self.__add_a_file(dsid, row, True))
                    else:
                        if 'attachments' in what_to_add:
                            new_csv_content.append(self.__add_an_attachment(dsid, row, parent=row['parents']))
                        if 'filesets' in what_to_add:
                            new_csv_content.append(self.__add_a_file(dsid, row))
            else:
                raise Exception(f"Unknown work type on {row['source_identifier']}: {row['model']}")
        return new_csv_content

    def write_csv(self, filename):
        with open(filename, 'w', newline='', encoding='utf-8') as bulkrax_sheet:
            writer = csv.DictWriter(bulkrax_sheet, fieldnames=self.headers)
            writer.writeheader()
            for data in self.new_csv_with_files:
                writer.writerow(data)
        return


class FileSetFinder:
    def __init__(self, pid):
        self.universal_ignores = ('DC', 'RELS-EXT', 'TECHMD', 'PREVIEW', 'JPG', 'JP2', 'MEDIUM_SIZE', 'POLICY', 'TN')
        self.pid = pid.replace('.xml', '')
        self.files = self.__get_all_files()

    def __get_all_files(self):
        results = ResourceIndexSearch().get_files(self.pid)
        return [result for result in results if result not in self.universal_ignores]


class RDFTypeGenerator:
    def __init__(self, parent_type):
        self.parent_type = parent_type

    def find_file_types(self, dsid, preserve_and_obj):
        if self.parent_type == "Image":
            return self.__get_rdf_types_for_file_on_an_image(dsid, preserve_and_obj)
        elif self.parent_type == "Audio":
            return self.__get_rdf_types_for_file_on_an_audio_work(dsid, preserve_and_obj)
        elif self.parent_type == "Video":
            return self.__get_rdf_types_for_file_on_a_video_work(dsid, preserve_and_obj)
        elif self.parent_type == "Book":
            return self.__get_rdf_types_for_file_on_a_book(dsid, preserve_and_obj)
        elif self.parent_type == "Pdf":
            return self.__get_rdf_types_for_file_on_a_pdf_work(dsid, preserve_and_obj)
        elif self.parent_type == "Page":
            return self.__get_rdf_types_for_file_on_a_pdf_work(dsid, preserve_and_obj)
        else:
            raise Exception(f"Parent type unknown: {self.parent_type}")

    @staticmethod
    def __get_rdf_types_for_file_on_an_image(dsid, preserve_and_obj):
        if dsid == "OBJ" and preserve_and_obj is False:
            return "http://pcdm.org/use#PreservationFile | http://pcdm.org/use#IntermediateFile"
        elif dsid == "OBJ":
            return "http://pcdm.org/use#IntermediateFile"
        elif dsid == "PRESERVE":
            return "http://pcdm.org/use#PreservationFile"
        elif dsid == "MODS":
            return "http://pcdm.org/file-format-types#Markup"
        elif dsid == "POLICY":
            return "http://pcdm.org/file-format-types#StructuredText"
        elif dsid == "OCR":
            return "http://pcdm.org/use#ExtractedText"
        elif dsid == "HOCR":
            return "http://pcdm.org/file-format-types#HTML"
        else:
            return "http://pcdm.org/use#OriginalFile"

    @staticmethod
    def __get_rdf_types_for_file_on_an_audio_work(dsid, preserve_and_obj):
        if dsid == "OBJ" and preserve_and_obj is False:
            return "http://pcdm.org/use#PreservationFile | http://pcdm.org/use#IntermediateFile"
        elif dsid == "OBJ":
            return "http://pcdm.org/use#PreservationFile"
        elif dsid == "PROXY_MP3":
            return "http://pcdm.org/use#IntermediateFile"
        elif dsid == "POLICY":
            return "http://pcdm.org/file-format-types#StructuredText"
        elif dsid == "MODS":
            return "http://pcdm.org/file-format-types#Markup"
        elif dsid == "TRANSCRIPT":
            return "http://pcdm.org/use#Transcript"
        elif dsid == "TRANSCRIPT-ES":
            return "http://pcdm.org/use#Transcript"
        elif dsid == "TN":
            return "http://pcdm.org/use#ThumbnailImage"
        else:
            return "http://pcdm.org/use#OriginalFile"

    @staticmethod
    def __get_rdf_types_for_file_on_a_video_work(dsid, preserve_and_obj):
        if dsid == "OBJ" and preserve_and_obj is False:
            return "http://pcdm.org/use#PreservationFile | http://pcdm.org/use#IntermediateFile"
        elif dsid == "OBJ":
            return "http://pcdm.org/use#PreservationFile"
        elif dsid == "MP4":
            return "http://pcdm.org/use#IntermediateFile"
        elif dsid == "POLICY":
            return "http://pcdm.org/file-format-types#StructuredText"
        elif dsid == "MODS":
            return "http://pcdm.org/file-format-types#Markup"
        elif dsid == "TRANSCRIPT":
            return "http://pcdm.org/use#Transcript"
        elif dsid == "TRANSCRIPT-ES":
            return "http://pcdm.org/use#Transcript"
        elif dsid == "TN":
            return "http://pcdm.org/use#ThumbnailImage"
        else:
            return "http://pcdm.org/use#OriginalFile"

    @staticmethod
    def __get_rdf_types_for_file_on_a_pdf_work(dsid, preserve_and_obj):
        if dsid == "OBJ" and preserve_and_obj is False:
            return "http://pcdm.org/use#PreservationFile | http://pcdm.org/use#IntermediateFile"
        elif dsid == "PDFA":
            return "http://pcdm.org/use#PreservationFile"
        elif dsid == "OBJ":
            return "http://pcdm.org/use#IntermediateFile | http://pcdm.org/use#OriginalFile"
        elif dsid == "POLICY":
            return "http://pcdm.org/file-format-types#StructuredText"
        elif dsid == "MODS":
            return "http://pcdm.org/file-format-types#Markup"
        else:
            return "http://pcdm.org/use#OriginalFile"

    @staticmethod
    def __get_rdf_types_for_file_on_a_book(dsid, preserve_and_obj):
        if dsid == "OBJ" and preserve_and_obj is False:
            return "http://pcdm.org/use#PreservationFile | http://pcdm.org/use#IntermediateFile"
        elif dsid == "OBJ":
            return "http://pcdm.org/use#PreservationFile"
        elif dsid == "MODS":
            return "http://pcdm.org/file-format-types#Markup"
        elif dsid == "TRANSCRIPT":
            return "http://pcdm.org/use#Transcript"
        elif dsid == "OCR":
            return "http://pcdm.org/use#Transcript"
        elif dsid == "PDF":
            return "http://pcdm.org/file-format-types#Document | http://pcdm.org/use#ServiceFile"
        elif dsid == "ORIGINAL":
            return "http://pcdm.org/file-format-types#Document | http://pcdm.org/use#ServiceFile | http://pcdm.org/use#OriginalFile"
        elif dsid == "ORIGINAL_EDITED":
            return "http://pcdm.org/file-format-types#Document | http://pcdm.org/use#ServiceFile"
        elif dsid == "TEI":
            return "http://pcdm.org/file-format-types#Markup"
        else:
            return "http://pcdm.org/use#OriginalFile"


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description='Add files to a CSV.')
    parser.add_argument("-s", "--sheet", dest="sheet", help="Specify original csv.", required=True)
    parser.add_argument(
        "-f", "--files_sheet", dest="files_sheet", help="Optional: specify files sheet."
    )
    parser.add_argument(
        '-w', "--what_to_add", dest="what_to_add", help="What To Add to Sheet", default="everything"
    )
    parser.add_argument(
        '-r', "--remote", dest="remote", help="Remote File Server", default='https://digital.lib.utk.edu/collections/islandora/object/'
    )
    args = parser.parse_args()
    files_sheet = f"{args.sheet.replace('.csv', '')}_with_filesets_and_attachments.csv"
    what_to_add = ['filesets', 'attachments']
    if args.files_sheet:
        files_sheet = args.files_sheet
    if args.what_to_add != 'everything':
        what_to_add = [args.what_to_add]
    """Take a CSV and Add files to it"""
    x = FileOrganizer(
        args.sheet,
        what_to_add,
        args.remote
    )
    x.write_csv(files_sheet)
    """Below: Get datastreams of a PID without the ones to ignore"""
    # x = FileSetFinder('heilman:150')
    # print(x.files)
    """Below: Get Large Images from a Collection with on constituent parts of a compound object"""
    # x = ResourceIndexSearch().get_images_no_parts('collections:boydcs')
    # with open('temp/things_to_download.txt', 'w') as things_to_download:
    #     for pid in x:
    #         things_to_download.write(f"{pid}\n")



import yaml
import os
import csv
from tqdm import tqdm
from .base import BaseProperty, StandardProperty, XMLtoDictProperty
from utk_exodus.risearch import ResourceIndexSearch


class TitleProperty(BaseProperty):
    """
    Used for Titles.

    Handles the mapping of different types of titles based on the provided XML data.

    - If just `titleInfo/title`, map to dcterms:title.
    - If just `titleInfo[@supplied="yes"]/title`, map to dcterms:title.
    - If `titleInfo[@supplied="yes"]/title` and `titleInfo/title`, map the former to dcterms:title, and the latter to
      dcterms:alternative.
    - If `titleInfo/partName`, concat to `titleInfo/title` in dcterms:title with a `,`.
    - If `titleInfo/partNumber`, concat to `titleInfo[@supplied="yes"]` in dcterms:alternative.
    - If `titleInfo/nonSort`, concat to `titleInfo/title` in dcterms:title.
    - If `titleInfo[@type="alternative"]`, map to dcterms:alternative.
    """

    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)
        self.various_titles = self.__extract_titles()

    def __extract_titles(self):
        titles_data = {
            "plain": self.__get_titles_from_xpath(
                'mods:titleInfo[not(@supplied)][not(@type="alternative")]/mods:title'
            ),
            "supplied": self.__get_titles_from_xpath(
                "mods:titleInfo[@supplied]/mods:title"
            ),
            "part_names": self.__get_titles_from_xpath("mods:titleInfo/mods:partName"),
            "part_numbers": self.__get_titles_from_xpath(
                "mods:titleInfo/mods:partNumber"
            ),
            "non_sorts": self.__get_titles_from_xpath("mods:titleInfo/mods:nonSort"),
            "alternatives": self.__get_titles_from_xpath(
                'mods:titleInfo[@type="alternative"]/mods:title'
            ),
        }
        return titles_data

    def __get_titles_from_xpath(self, xpath_expr):
        # Retrieve text content based on the given xpath expression.
        return [
            element.text
            for element in self.root.xpath(xpath_expr, namespaces=self.namespaces)
        ]

    def find(self):
        """
        Handles the mapping of different types of titles based on the provided XML data.

        Returns:
            dict: A dictionary containing the mapped titles.

        Example:
             >>> titles = TitleProperty("tests/fixtures/utsmc_17870.xml", {"mods": "http://www.loc.gov/mods/v3"})
             >>> titles.find()
             ... #doctest: +NORMALIZE_WHITESPACE
             {'title': ['Prussian heroes march'], 'alternative_title': ['Prussian heroes: Prussen helden march']}
        """
        titles = []
        alternatives = []
        titles_data = self.various_titles

        # Handle supplied and plain titles mapping to dcterms:title and dcterms:alternative.
        if titles_data["supplied"] and titles_data["plain"]:
            titles.extend(titles_data["supplied"])
            alternatives.extend(titles_data["plain"])
        else:
            titles.extend(titles_data["supplied"] + titles_data["plain"])

        # Handle alternative titles mapping to dcterms:alternative.
        alternatives.extend(titles_data["alternatives"])

        return {"title": titles, "alternative_title": alternatives}


class RoleAndNameProperty(XMLtoDictProperty):
    def __init__(self, file):
        super().__init__(file)
        self.all_names = self.__find_all_names()

    def __find_all_names(self):
        if "mods:name" in self.doc["mods:mods"]:
            all_names = self.doc["mods:mods"]["mods:name"]
            if isinstance(all_names, list):
                return all_names
            elif isinstance(all_names, dict):
                return [all_names]
            elif isinstance(all_names, str):
                return [all_names]
            else:
                return ["Problem"]
        else:
            return []

    def find(self):
        """
        Find all names and roles in the XML file.

        Returns:
            dict: A dictionary containing the roles and names.

        Examples:
            >>> roles_and_names = RoleAndNameProperty("tests/fixtures/harp_1.xml")
            >>> roles_and_names.find()
            {'utk_composer': ['Swan, W. H. (William H.)', 'Swan, Marcus Lafayette'], 'utk_compiler': ['Swan, W. H. (William H.)', 'Swan, Marcus Lafayette']}
        """
        roles_and_names = {}
        for name in self.all_names:
            local_roles = []
            try:
                local_roles.append(
                    f"utk_{name['mods:role']['mods:roleTerm']['#text'].lower().replace(' ', '_')}"
                )
            except KeyError:
                print(name)
            # TODO: A name can have multiple roles
            except TypeError:
                for role in name["mods:role"]:
                    local_roles.append(
                        f"utk_{role['mods:roleTerm']['#text'].lower().replace(' ', '_')}"
                    )
            # TODO: Rework this.  It's not pretty but it works.
            name_value = name["mods:namePart"]
            for role in local_roles:
                if isinstance(name_value, list):
                    for part in name_value:
                        if isinstance(part, dict) and part.get("mods:namePart"):
                            if role not in roles_and_names:
                                roles_and_names[role] = [part["mods:namePart"]]
                            else:
                                roles_and_names[role].append([part["mods:namePart"]])
                        if isinstance(part, str) and not part.startswith("http"):
                            if role not in roles_and_names:
                                roles_and_names[role] = [part]
                            else:
                                roles_and_names[role].append(part)
                elif role not in roles_and_names and not name_value.startswith("http"):
                    roles_and_names[role] = [name_value]
                elif not name_value.startswith("http"):
                    roles_and_names[role].append(name_value)
        return roles_and_names


class NameProperty(XMLtoDictProperty):
    """
    Used for names.
    """

    def __init__(self, file):
        super().__init__(file)
        self.all_names = self.__find_all_names()

    def __find_all_names(self):
        if "mods:name" in self.doc["mods:mods"]:
            all_names = self.doc["mods:mods"]["mods:name"]
            if isinstance(all_names, list):
                return all_names
            elif isinstance(all_names, dict):
                return [all_names]
            elif isinstance(all_names, str):
                return [all_names]
            else:
                return ["Problem"]
        else:
            return []

    def find(self):
        """
        Find all names in the XML file.

        Returns:
            dict: A dictionary containing the names.

        Examples:
            >>> names = NameProperty("tests/fixtures/harp_1.xml")
            >>> names.find()
            {'composer': ['http://id.loc.gov/authorities/names/no2002022963', 'http://id.loc.gov/authorities/names/n78013127'], 'compiler': ['http://id.loc.gov/authorities/names/no2002022963', 'http://id.loc.gov/authorities/names/n78013127']}
        """
        roles_and_names = {}
        for name in self.all_names:
            roles = []
            local_roles = []
            try:
                roles.append(
                    name["mods:role"]["mods:roleTerm"]["#text"]
                    .lower()
                    .replace(" ", "_")
                )
                local_roles.append(
                    f"utk_{name['mods:role']['mods:roleTerm']['#text'].lower().replace(' ', '_')}"
                )
            except KeyError:
                print(name)
            # TODO: A name can have multiple roles
            except TypeError:
                if isinstance(name["mods:role"], list):
                    for role in name["mods:role"]:
                        if "#text" in role["mods:roleTerm"]:
                            roles.append(
                                role["mods:roleTerm"]["#text"].lower().replace(" ", "_")
                            )
                            local_roles.append(
                                f"utk_{role['mods:roleTerm']['#text'].lower().replace(' ', '_')}"
                            )
                        else:
                            roles.append(
                                role["mods:roleTerm"].lower().replace(" ", "_")
                            )
                            local_roles.append(
                                f"utk_{role['mods:roleTerm'].lower().replace(' ', '_')}"
                            )
                else:
                    roles.append(
                        name["mods:role"]["mods:roleTerm"].lower().replace(" ", "_")
                    )
                    local_roles.append(
                        f"utk_{name['mods:role']['mods:roleTerm'].lower().replace(' ', '_')}"
                    )
            # TODO: Rework this.  It's not pretty but it works.
            name_value = name["mods:namePart"]
            if "@valueURI" in name:
                name_value = name["@valueURI"]
            for role in roles:
                if isinstance(name_value, list):
                    for part in name_value:
                        if isinstance(part, dict) and "@valueURI" in part:
                            if role not in roles_and_names:
                                roles_and_names[role] = [part["@valueURI"]]
                            else:
                                roles_and_names[role].append([part["@valueURI"]])
                elif role not in roles_and_names and name_value.startswith("http"):
                    roles_and_names[role] = [name_value]
                elif name_value.startswith("http"):
                    roles_and_names[role].append(name_value)
            for role in local_roles:
                if isinstance(name_value, list):
                    for part in name_value:
                        if isinstance(part, str) and not part.startswith("http"):
                            if role not in roles_and_names:
                                roles_and_names[role] = [part]
                            else:
                                roles_and_names[role].append(part)
                elif role not in roles_and_names and not name_value.startswith("http"):
                    roles_and_names[role] = [name_value]
                elif not name_value.startswith("http"):
                    roles_and_names[role].append(name_value)
        return roles_and_names


class GeoNamesProperty(BaseProperty):
    """
    Handles the mapping of different types of geonames based on the provided XML data.

    Args:
        path (str): The path to the XML file.
        namespaces (dict): A dictionary containing the namespaces used in the XML file.
    """

    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find(self, name):
        """
        Find all geonames in the XML file.

        Args:
            name (str): The name of the geoname.

        Returns:
            dict: A dictionary containing the geoname mappings.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> geonames = GeoNamesProperty("tests/fixtures/webster_1127.xml", NAMESPACES)
            >>> geonames.find("spatial")
            {'spatial': ['http://sws.geonames.org/4050810', 'http://sws.geonames.org/4609260', 'http://id.loc.gov/authorities/subjects/sh85057008']}
        """
        uris = [
            uri.replace("about.rdf", "")
            for uri in self.root.xpath(
                "mods:subject/mods:geographic/@valueURI", namespaces=self.namespaces
            )
        ]
        lc_uris = [
            uri
            for uri in self.root.xpath(
                "mods:subject[mods:geographic]/@valueURI", namespaces=self.namespaces
            )
        ]
        all_values = []
        for uri in lc_uris:
            all_values.append(uri)
        for uri in uris:
            all_values.append(uri)
        return {name: all_values}


class PhysicalLocationsProperties(BaseProperty):
    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def __find_repositories(self):
        with_repository_designation = [
            thing.text
            for thing in self.root.xpath(
                'mods:location/mods:physicalLocation[@displayLabel="Repository"]',
                namespaces=self.namespaces,
            )
        ]
        others = [
            thing.text
            for thing in self.root.xpath(
                "mods:location/mods:physicalLocation[not(@displayLabel)]",
                namespaces=self.namespaces,
            )
        ]
        all_repositories = []
        for value in with_repository_designation:
            all_repositories.append(value)
        for value in others:
            if value is None or "University of Tennesse" in value:
                all_repositories.append(
                    "University of Tennessee, Knoxville. Special Collections"
                )
            else:
                all_repositories.append(value)
        return all_repositories

    def find(self):
        """
        Find all locations properties in the XML file.

        Returns:
            dict: A dictionary containing location properties.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> physical_location = PhysicalLocationsProperties("tests/fixtures/civilwar_1438.xml", NAMESPACES)
            >>> physical_location.find()
            {'repository': ['University of Tennessee, Knoxville. Special Collections'], 'archival_collection': ['O. P. Temple Papers,1862']}
        """
        return {
            "repository": self.__find_repositories(),
            "archival_collection": self.__find_archival_collections(),
        }

    def __find_archival_collections(self):
        all_archival_collections = []
        other_archival_collections = [
            collection.text
            for collection in self.root.xpath(
                'mods:location/mods:physicalLocation[@displayLabel="Collection"]',
                namespaces=self.namespaces,
            )
        ]
        primary_archival_collections = [
            collection
            for collection in self.root.xpath(
                'mods:relatedItem[@displayLabel="Collection"][mods:titleInfo]',
                namespaces=self.namespaces,
            )
        ]
        for collection in other_archival_collections:
            if collection not in all_archival_collections:
                all_archival_collections.append(collection)
        for collection in primary_archival_collections:
            identifier = ""
            title = ""
            output = ""
            for child in collection:
                if child.tag == "{http://www.loc.gov/mods/v3}identifier":
                    identifier = child.text
                elif child.tag == "{http://www.loc.gov/mods/v3}titleInfo":
                    for sub_child in child:
                        if sub_child.tag == "{http://www.loc.gov/mods/v3}title":
                            title = sub_child.text
            if title != "" and identifier != "":
                output = f"{title}, {identifier}"
            elif identifier != "":
                output = identifier
            elif title != "":
                output = title
            if output != "":
                all_archival_collections.append(output)
        return all_archival_collections


class DataProvider(BaseProperty):
    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find(self):
        """
        Find all data providers in the XML file.

        Returns:
            dict: A dictionary containing data provider information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> dataProvider = DataProvider("tests/fixtures/egypt_224.xml", NAMESPACES)
            >>> dataProvider.find()
            {'provider': ['University of Tennessee, Knoxville. Libraries'], 'intermediate_provider': ['Frank H. McClung Museum of Natural History and Culture']}
        """
        values = [
            value.text
            for value in self.root.xpath(
                "mods:recordInfo/mods:recordContentSource", namespaces=self.namespaces
            )
        ]
        return {
            "provider": ["University of Tennessee, Knoxville. Libraries"],
            "intermediate_provider": [
                value
                for value in values
                if value != "University of Tennessee, Knoxville. Libraries"
            ],
        }


class MachineDate(BaseProperty):
    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find(self):
        """
        Find all machine date information in the XML file.

        Returns:
            dict: A dictionary containing machine date information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> machineDate = MachineDate("tests/fixtures/volvoices_2993.xml", NAMESPACES)
            >>> machineDate.find()
            {'date_created_d': ['1948-01'], 'date_issued_d': ['1948'], 'date_other_d': []}
        """
        date_created = [
            value.text
            for value in self.root.xpath(
                'mods:originInfo/mods:dateCreated[@encoding="edtf"]',
                namespaces=self.namespaces,
            )
        ]
        date_issued = [
            value.text
            for value in self.root.xpath(
                'mods:originInfo/mods:dateIssued[@encoding="edtf"]',
                namespaces=self.namespaces,
            )
        ]
        date_other = [
            value.text
            for value in self.root.xpath(
                'mods:originInfo/mods:dateOther[@encoding="edtf"]',
                namespaces=self.namespaces,
            )
        ]
        return {
            "date_created_d": self.__sort_if_range(date_created),
            "date_issued_d": self.__sort_if_range(date_issued),
            "date_other_d": self.__sort_if_range(date_other),
        }

    @staticmethod
    def __sort_if_range(values):
        if len(values) == 2 and None not in values:
            values.sort()
            return [f"{values[0]}/{values[1]}"]
        elif len(values) == 2:
            for value in values:
                if value is not None:
                    return [value]
        else:
            return values


class SubjectProperty(BaseProperty):
    """Get subjects from the MODS XML file."""

    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find_topic(self):
        """Find all topics in the XML file.

        Returns:
            dict: A dictionary containing topics information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> subjects = SubjectProperty("tests/fixtures/knoxgardens_125.xml", NAMESPACES)
            >>> subjects.find_topic()
            {'subject': ['http://id.loc.gov/authorities/subjects/sh85101348', 'http://id.loc.gov/authorities/subjects/sh85053123', 'http://id.loc.gov/authorities/subjects/sh85103022', 'http://id.loc.gov/authorities/subjects/sh2008120720']}

        """
        xpaths = [
            "mods:subject[mods:topic]/@valueURI",
            "mods:subject/mods:topic/@valueURI",
            "mods:subject[mods:name/mods:namePart]/@valueURI",
            "mods:subject/mods:name/@valueURI",
            'mods:genre[@authority="aat"]/@valueURI',
            'mods:genre[@authority="lcmpt"]/@valueURI',
            'mods:genre[@authority="lcsh"]/@valueURI',
        ]

        # Execute each XPath query and collect results
        return_values = []
        for xpath in xpaths:
            uris = self.root.xpath(xpath, namespaces=self.namespaces)
            return_values.extend(uri.strip() for uri in uris)

        return {"subject": return_values}


class KeywordProperty(BaseProperty):
    """Get keywords from the MODS XML file."""

    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find_topic(self):
        """Find all topics in the XML file.

        Returns:
            dict: A dictionary containing topics information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> keywords = KeywordProperty("tests/fixtures/civilwar_1438.xml", NAMESPACES)
            >>> keywords.find_topic()
            {'keyword': ['Jurisdiction -- Tennessee, East -- History -- Civil War, 1861-1865', 'Actions and defenses -- Tennessee, East -- History -- Civil War, 1861-1865', 'Tennessee, East -- Politics and government -- 19th century', 'Wallace, Jesse G. -- Correspondence', 'Temple, Oliver Perry, 1820-1907 -- Correspondence']}

        """
        non_uris_topics = [
            value.text
            for value in self.root.xpath(
                "mods:subject[not(@valueURI)]/mods:topic[not(@valueURI)]",
                namespaces=self.namespaces,
            )
        ]
        non_uris_names = [
            value.text
            for value in self.root.xpath(
                "mods:subject[not(@valueURI)]/mods:name[not(@valueURI)]/mods:namePart[not(@valueURI)]",
                namespaces=self.namespaces,
            )
        ]
        all_initial_values = [non_uris_topics, non_uris_names]
        return_values = []
        for iterable in all_initial_values:
            for value in iterable:
                return_values.append(value)
        return {"keyword": return_values}


class TypesProperties(BaseProperty):
    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find(self):
        """
        Find all types properties information in the XML file.

        Returns:
            dict: A dictionary containing types properties information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> types = TypesProperties("tests/fixtures/utsmc_17870.xml", NAMESPACES)
            >>> types.find()
            {'form': ['http://vocab.getty.edu/aat/300026430'], 'resource_type': ['http://id.loc.gov/vocabulary/resourceTypes/not', 'http://id.loc.gov/vocabulary/resourceTypes/txt'], 'form_local': []}
        """
        return {
            "form": self.__find_edm_has_type(),
            "resource_type": self.__find_dcterms_type(),
            "form_local": self.__find_local_form(),
        }

    def __find_dcterms_type(self):
        genre_uris = {
            "cartographic": "http://id.loc.gov/vocabulary/resourceTypes/car",
            "image": "http://id.loc.gov/vocabulary/resourceTypes/img",
            "notated music": "http://id.loc.gov/vocabulary/resourceTypes/not",
            "still image": "http://id.loc.gov/vocabulary/resourceTypes/img",
            "text": "http://id.loc.gov/vocabulary/resourceTypes/txt",
        }
        type_of_resource_uris = {
            "text": "http://id.loc.gov/vocabulary/resourceTypes/txt",
            "cartographic": "http://id.loc.gov/vocabulary/resourceTypes/car",
            "notated music": "http://id.loc.gov/vocabulary/resourceTypes/not",
            "sound recording-nonmusical": "http://id.loc.gov/vocabulary/resourceTypes/aun",
            "sound recording": "http://id.loc.gov/vocabulary/resourceTypes/aud",
            "still image": "http://id.loc.gov/vocabulary/resourceTypes/img",
            "moving image": "http://id.loc.gov/vocabulary/resourceTypes/mov",
            "three dimensional object": "http://id.loc.gov/vocabulary/resourceTypes/art",
        }
        # TODO: Works but messy!
        genre_to_dcterms_match_1 = [
            value.text
            for value in self.root.xpath(
                "mods:genre[not(@*)][string() = 'cartographic']",
                namespaces=self.namespaces,
            )
        ]
        genre_to_dcterms_match_2 = [
            value.text
            for value in self.root.xpath(
                "mods:genre[not(@*)][string() = 'notated music']",
                namespaces=self.namespaces,
            )
        ]
        genre_to_dcterms_match_3 = [
            value.text
            for value in self.root.xpath(
                "mods:genre[@authority = 'dct'][string() = 'image']",
                namespaces=self.namespaces,
            )
        ]
        genre_to_dcterms_match_4 = [
            value.text
            for value in self.root.xpath(
                "mods:genre[@authority = 'dct'][string() = 'still image']",
                namespaces=self.namespaces,
            )
        ]
        genre_to_dcterms_match_5 = [
            value.text
            for value in self.root.xpath(
                "mods:genre[@authority = 'dct'][string() = 'text']",
                namespaces=self.namespaces,
            )
        ]
        genre_to_dcterms_matches = (
            genre_to_dcterms_match_1,
            genre_to_dcterms_match_2,
            genre_to_dcterms_match_3,
            genre_to_dcterms_match_4,
            genre_to_dcterms_match_5,
        )
        type_of_resource_to_dcterms_type = [
            value.text
            for value in self.root.xpath(
                "mods:typeOfResource[not(@collection)]", namespaces=self.namespaces
            )
        ]
        type_of_resource_collection = [
            value.text
            for value in self.root.xpath(
                "mods:typeOfResource[@collection]", namespaces=self.namespaces
            )
        ]
        all_dcterms_types = []
        for matches in genre_to_dcterms_matches:
            for match in matches:
                if match in genre_uris:
                    all_dcterms_types.append(genre_uris[match])
        for value in type_of_resource_to_dcterms_type:
            if value in type_of_resource_uris:
                all_dcterms_types.append(type_of_resource_uris[value])

        if len(type_of_resource_collection) > 0:
            all_dcterms_types.append("http://id.loc.gov/vocabulary/resourceTypes/col")
        return all_dcterms_types

    def __find_edm_has_type(self):
        lcgft_genres = [
            uri
            for uri in self.root.xpath(
                'mods:genre[@authority="lcgft"]/@valueURI', namespaces=self.namespaces
            )
        ]
        form_uris = [
            uri
            for uri in self.root.xpath(
                "mods:physicalDescription/mods:form/@valueURI",
                namespaces=self.namespaces,
            )
        ]
        all_matches = [lcgft_genres, form_uris]
        return_values = []
        for match in all_matches:
            for value in match:
                return_values.append(value)
        return return_values

    def __find_local_form(self):
        form_no_uri = [
            value.text
            for value in self.root.xpath(
                "mods:physicalDescription/mods:form[not(@valueURI)][not(@type='material')]",
                namespaces=self.namespaces,
            )
        ]
        genre_strings = [
            value.text
            for value in self.root.xpath(
                "mods:genre[not(@*) and not(text()='cartographic') and not(text()='notated music')]",
                namespaces=self.namespaces,
            )
        ]
        all_matches = [form_no_uri, genre_strings]
        return_values = []
        for match in all_matches:
            for value in match:
                return_values.append(value)
        return return_values


class LocalTypesProperties(BaseProperty):
    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find(self):
        """
        Find all local types properties information in the XML file.

        Returns:
            dict: A dictionary containing local types properties information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> local_types = LocalTypesProperties("tests/fixtures/egypt_224.xml", NAMESPACES)
            >>> local_types.find()
            {'resource_type_local': ['still image'], 'form_local': ['platinum prints']}
        """
        return {
            "resource_type_local": self.__find_dcterms_type(),
            "form_local": self.__find_local_form(),
        }

    def __find_dcterms_type(self):
        genre_uris = {
            "cartographic": "http://id.loc.gov/vocabulary/resourceTypes/car",
            "image": "http://id.loc.gov/vocabulary/resourceTypes/img",
            "notated music": "http://id.loc.gov/vocabulary/resourceTypes/not",
            "still image": "http://id.loc.gov/vocabulary/resourceTypes/img",
            "text": "http://id.loc.gov/vocabulary/resourceTypes/txt",
        }
        type_of_resource_uris = {
            "text": "http://id.loc.gov/vocabulary/resourceTypes/txt",
            "cartographic": "http://id.loc.gov/vocabulary/resourceTypes/car",
            "notated music": "http://id.loc.gov/vocabulary/resourceTypes/not",
            "sound recording-nonmusical": "http://id.loc.gov/vocabulary/resourceTypes/aun",
            "sound recording": "http://id.loc.gov/vocabulary/resourceTypes/aud",
            "still image": "http://id.loc.gov/vocabulary/resourceTypes/img",
            "moving image": "http://id.loc.gov/vocabulary/resourceTypes/mov",
            "three dimensional object": "http://id.loc.gov/vocabulary/resourceTypes/art",
        }
        # TODO: Works but messy!
        genre_to_dcterms_match_1 = [
            value.text
            for value in self.root.xpath(
                "mods:genre[not(@*)][string() = 'cartographic']",
                namespaces=self.namespaces,
            )
        ]
        genre_to_dcterms_match_2 = [
            value.text
            for value in self.root.xpath(
                "mods:genre[not(@*)][string() = 'notated music']",
                namespaces=self.namespaces,
            )
        ]
        genre_to_dcterms_match_3 = [
            value.text
            for value in self.root.xpath(
                "mods:genre[@authority = 'dct'][string() = 'image']",
                namespaces=self.namespaces,
            )
        ]
        genre_to_dcterms_match_4 = [
            value.text
            for value in self.root.xpath(
                "mods:genre[@authority = 'dct'][string() = 'still image']",
                namespaces=self.namespaces,
            )
        ]
        genre_to_dcterms_match_5 = [
            value.text
            for value in self.root.xpath(
                "mods:genre[@authority = 'dct'][string() = 'text']",
                namespaces=self.namespaces,
            )
        ]
        genre_to_dcterms_matches = (
            genre_to_dcterms_match_1,
            genre_to_dcterms_match_2,
            genre_to_dcterms_match_3,
            genre_to_dcterms_match_4,
            genre_to_dcterms_match_5,
        )
        type_of_resource_to_dcterms_type = [
            value.text
            for value in self.root.xpath(
                "mods:typeOfResource[not(@collection)]", namespaces=self.namespaces
            )
        ]
        type_of_resource_collection = [
            value.text
            for value in self.root.xpath(
                "mods:typeOfResource[@collection]", namespaces=self.namespaces
            )
        ]
        all_dcterms_types = []
        for matches in genre_to_dcterms_matches:
            for match in matches:
                if match in genre_uris:
                    all_dcterms_types.append(match)
        for value in type_of_resource_to_dcterms_type:
            if value in type_of_resource_uris:
                all_dcterms_types.append(value)

        if len(type_of_resource_collection) > 0:
            all_dcterms_types.append("http://id.loc.gov/vocabulary/resourceTypes/col")
        return all_dcterms_types

    def __find_local_form(self):
        form_no_uri = [
            value.text
            for value in self.root.xpath(
                "mods:physicalDescription/mods:form[not(@type='material')]",
                namespaces=self.namespaces,
            )
        ]
        genre_strings = [
            value.text
            for value in self.root.xpath(
                "mods:genre[not(text()='cartographic') and not(text()='notated music')]",
                namespaces=self.namespaces,
            )
        ]
        all_matches = [form_no_uri, genre_strings]
        return_values = []
        for match in all_matches:
            for value in match:
                return_values.append(value)
        return return_values


class PublisherProperty(BaseProperty):
    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find(self):
        """
        Find all publisher property information in the XML file.

        Returns:
            dict: A dictionary containing publisher property information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> publisher = PublisherProperty("tests/fixtures/playbills:1052.xml", NAMESPACES)
            >>> publisher.find()
            {'publisher': []}
        """
        return {
            "publisher": [
                uri
                for uri in self.root.xpath(
                    "mods:originInfo/mods:publisher/@valueURI",
                    namespaces=self.namespaces,
                )
            ]
        }


class RightsOrLicenseProperties(BaseProperty):
    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find(self):
        """
        Find all rights or liscense property information in the XML file.

        Returns:
            dict: A dictionary containing rights or license property information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> rights = RightsOrLicenseProperties("tests/fixtures/heilman:1010.xml", NAMESPACES)
            >>> rights.find()
            {'license': ['http://creativecommons.org/licenses/by-nc-nd/3.0/'], 'rights_statement': ['http://rightsstatements.org/vocab/InC/1.0/']}
        """
        final = {}
        rights = [
            uri
            for uri in self.root.xpath(
                'mods:accessCondition[not(@type="restriction on access")]/@xlink:href',
                namespaces=self.namespaces,
            )
            if "rightsstatements.org" in uri
        ]
        licenses = [
            uri.replace("https://", "http://")
            for uri in self.root.xpath(
                'mods:accessCondition[not(@type="restriction on access")]/@xlink:href',
                namespaces=self.namespaces,
            )
            if "creativecommons.org" in uri
        ]
        if len(rights) > 0:
            final["rights_statement"] = rights
        if len(licenses) > 0:
            final["license"] = licenses
        if len(rights) == 0:
            if licenses[0] == "http://creativecommons.org/publicdomain/zero/1.0/":
                final["rights_statement"] = [
                    "http://rightsstatements.org/vocab/NKC/1.0/"
                ]
            else:
                final["rights_statement"] = [
                    "http://rightsstatements.org/vocab/InC/1.0/"
                ]
        return final


class PublicationPlaceProperty(BaseProperty):
    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find(self):
        """
        Find all publication place property information in the XML file.

        Returns:
            dict: A dictionary containing publication place property information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> publication_place = PublicationPlaceProperty("tests/fixtures/volvoices_2495.xml", NAMESPACES)
            >>> publication_place.find()
            {'publication_place': ['http://id.loc.gov/authorities/names/n82063401']}
        """
        return {
            "publication_place": [
                uri
                for uri in self.root.xpath(
                    "mods:originInfo/mods:place/mods:placeTerm/@valueURI",
                    namespaces=self.namespaces,
                )
            ]
        }


class LanguageURIProperty(BaseProperty):
    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find_term(self):
        """
        Find all language URI properties in the XML file.

        Returns:
            dict: A dictionary containing language URI property information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> language_uri = LanguageURIProperty("tests/fixtures/utsmc:725.xml", NAMESPACES)
            >>> language_uri.find_term()
            {'language': ['http://id.loc.gov/vocabulary/iso639-2/fre', 'http://id.loc.gov/vocabulary/iso639-2/ita']}
        """
        terms_and_uris = {
            "English": "http://id.loc.gov/vocabulary/iso639-2/eng",
            "French": "http://id.loc.gov/vocabulary/iso639-2/fre",
            "German": "http://id.loc.gov/vocabulary/iso639-2/ger",
            "Italian": "http://id.loc.gov/vocabulary/iso639-2/ita",
            "Latin": "http://id.loc.gov/vocabulary/iso639-2/lat",
            "No linguistic content": "http://id.loc.gov/vocabulary/iso639-2/zxx",
            "Russian": "http://id.loc.gov/vocabulary/iso639-2/rus",
            "Spanish": "http://id.loc.gov/vocabulary/iso639-2/spa",
            "Swedish": "http://id.loc.gov/vocabulary/iso639-2/swe",
            "en": "http://id.loc.gov/vocabulary/iso639-2/eng",
        }
        language_terms = [
            value.text
            for value in self.root.xpath(
                "mods:language/mods:languageTerm", namespaces=self.namespaces
            )
        ]
        lanuage_uris = []
        for language in language_terms:
            if language in terms_and_uris:
                lanuage_uris.append(terms_and_uris[language])
        return {"language": lanuage_uris}


class ExtentProperty(BaseProperty):
    """
    Class to Handle Extent Information.

    Args:
        path (str): The path to the file.
        namespaces (dict): Namespaces to be used in the XPath queries.
    """

    def __init__(self, path: str, namespaces: dict):
        super().__init__(path, namespaces)

    def find(self):
        """
        Finds and returns a dictionary containing the extent information of a record.

        Returns:
            dict: A dictionary containing the extent information.

        Examples:
            >>> NAMESPACES = {
            ... 'mods': 'http://www.loc.gov/mods/v3',
            ... 'xlink': 'http://www.w3.org/1999/xlink' }
            >>> extent_property = ExtentProperty("tests/fixtures/knoxgardens_125.xml", NAMESPACES)
            >>> extent_property.find()
            {'extent': ['3 1/4 x 5 inches']}

        """
        extents_without_units = [
            text.text
            for text in self.root.xpath(
                "mods:physicalDescription/mods:extent[not(@unit)]",
                namespaces=self.namespaces,
            )
        ]

        extents_with_units = [
            f"{node.text} {node.attrib['unit']}"
            for node in self.root.xpath(
                "mods:physicalDescription/mods:extent[@unit]",
                namespaces=self.namespaces,
            )
            if node.text is not None and "unit" in node.attrib
        ]

        # Combine extents with and without units into a single list
        final_extent = extents_with_units + extents_without_units

        return {"extent": final_extent}


class MetadataMapping:
    def __init__(self, path_to_mapping, file_path, membership_details=None):
        self.path = path_to_mapping
        self.membership_details = membership_details
        self.fieldnames = []
        self.all_files = self.__get_all_files(file_path)
        self.mapping_data = yaml.safe_load(open(path_to_mapping, "r"))["mapping"]
        self.namespaces = {
            "mods": "http://www.loc.gov/mods/v3",
            "xlink": "http://www.w3.org/1999/xlink",
        }
        self.output_data = self.__execute(self.namespaces)

    @staticmethod
    def __get_all_files(path):
        all_files = []
        for root, dirs, files in os.walk(path, topdown=False):
            for name in files:
                all_files.append(os.path.join(root, name))
        return all_files

    def __execute(self, namespaces):
        all_file_data = []
        all_pages = []
        for file in tqdm(self.all_files):
            # TODO: Ultimately, parents should be populated based on relationship.
            model = self.__dereference_islandora_type(file)
            output_data = {
                "source_identifier": file.split("/")[-1]
                .replace("_MODS.xml", "")
                .replace(".xml", ""),
                "model": model,
                "sequence": "",
                "remote_files": "",
                "parents": " | ".join(
                    ResourceIndexSearch().get_parent_collections(
                        file.split("/")[-1].replace("_MODS.xml", "").replace(".xml", "")
                    ),
                ),
                "has_work_type": self.__get_utk_ontology_value(model),
                "primary_identifier": file.split("/")[-1]
                .replace("_MODS.xml", "")
                .replace(".xml", ""),
            }
            self.__dereference_islandora_type(file)
            for rdf_property in self.mapping_data:
                if "special" not in rdf_property:
                    final_values = ""
                    values = StandardProperty(file, namespaces).find(
                        rdf_property["xpaths"]
                    )
                    if len(values) > 0:
                        # TODO: Make delimeter configurable
                        final_values = " | ".join(values)
                    output_data[rdf_property["name"]] = final_values
                else:
                    special = self.__lookup_special_property(
                        rdf_property["special"], file, namespaces, rdf_property["name"]
                    )
                    for k, v in special.items():
                        # TODO: Make delimeter configurable
                        if v != [[]]:
                            try:
                                output_data[k] = " | ".join(v)
                            except TypeError:
                                print(f"{TypeError}: {file}")
            self.__find_unique_fieldnames(output_data)
            all_file_data.append(output_data)
        for item in all_file_data:
            pages = self.look_for_pages(item)
            for page in pages:
                new_page = item.copy()
                new_page["source_identifier"] = page["pid"].replace("info:fedora/", "")
                new_page["parents"] = item["source_identifier"]
                new_page["model"] = "Page"
                new_page["sequence"] = page["page"]
                all_pages.append(new_page)
        for page in all_pages:
            all_file_data.append(page)
        return all_file_data

    def look_for_pages(self, data):
        if data["model"] == "Book":
            return ResourceIndexSearch().find_pages_in_book(data["source_identifier"])
        return []

    def __find_unique_fieldnames(self, data):
        for k, v in data.items():
            if k not in self.fieldnames:
                self.fieldnames.append(k)
        return

    def __dereference_islandora_type(self, file):
        islandora_types = {
            "info:fedora/islandora:sp-audioCModel": "Audio",
            "info:fedora/islandora:bookCModel": "Book",
            "info:fedora/islandora:binaryObjectCModel": "Generic",
            "info:fedora/islandora:sp_large_image_cmodel": "Image",
            "info:fedora/islandora:sp_basic_image": "Image",
            "info:fedora/islandora:sp_pdf": "Pdf",
            "info:fedora/islandora:sp_videoCModel": "Video",
        }
        x = ResourceIndexSearch().get_islandora_work_type(
            file.split("/")[-1]
            .replace("_MODS.xml", "")
            .replace(".xml", "")
            .replace("_", ":")
        )
        return islandora_types[x]

    @staticmethod
    def __get_utk_ontology_value(model):
        ontology_values = {
            "Audio": "https://ontology.lib.utk.edu/works#AudioWork",
            "Book": "https://ontology.lib.utk.edu/works#BookWork",
            "Generic": "https://ontology.lib.utk.edu/works#GenericWork",
            "Image": "https://ontology.lib.utk.edu/works#ImageWork",
            "Pdf": "https://ontology.lib.utk.edu/works#PDFWork",
            "Video": "https://ontology.lib.utk.edu/works#VideoWork",
        }
        return ontology_values[model]

    @staticmethod
    def __lookup_special_property(special_property, file, namespaces, name):
        if special_property == "TitleProperty":
            return TitleProperty(file, namespaces).find()
        elif special_property == "NameProperty":
            return NameProperty(file).find()
        elif special_property == "RoleAndNameProperty":
            return RoleAndNameProperty(file).find()
        elif special_property == "GeoNamesProperty":
            return GeoNamesProperty(file, namespaces).find(name)
        elif special_property == "DataProvider":
            return DataProvider(file, namespaces).find()
        elif special_property == "PhysicalLocationsProperties":
            return PhysicalLocationsProperties(file, namespaces).find()
        elif special_property == "SubjectProperty":
            return SubjectProperty(file, namespaces).find_topic()
        elif special_property == "KeywordProperty":
            return KeywordProperty(file, namespaces).find_topic()
        elif special_property == "TypesProperties":
            return TypesProperties(file, namespaces).find()
        elif special_property == "LocalTypesProperties":
            return LocalTypesProperties(file, namespaces).find()
        elif special_property == "LanguageURIProperty":
            return LanguageURIProperty(file, namespaces).find_term()
        elif special_property == "PublisherProperty":
            return PublisherProperty(file, namespaces).find()
        elif special_property == "PublicationPlaceProperty":
            return PublicationPlaceProperty(file, namespaces).find()
        elif special_property == "RightsOrLicenseProperties":
            return RightsOrLicenseProperties(file, namespaces).find()
        elif special_property == "ExtentProperty":
            return ExtentProperty(file, namespaces).find()
        elif special_property == "MachineDate":
            return MachineDate(file, namespaces).find()
        else:
            # Handle unknown special property
            raise ValueError(f"Unknown special property: {special_property}")

    def write_csv(self, filename):
        with open(filename, "w", newline="") as bulkrax_sheet:
            writer = csv.DictWriter(
                bulkrax_sheet, fieldnames=self.fieldnames, quoting=csv.QUOTE_MINIMAL
            )
            writer.writeheader()
            for data in self.output_data:
                writer.writerow(data)
        return



import csv


class ImportRefactor:
    """Remove old values from an import sheet."""

    def __init__(self, import_sheet, old_sheet):
        self.import_sheet = import_sheet
        self.old_sheet = old_sheet
        self.current_headers = self.__find_headers(import_sheet)
        self.old_headers = self.__find_headers(self.old_sheet)
        self.fields_to_nuke = list(set(self.old_headers) - set(self.current_headers))

    @staticmethod
    def __find_headers(sheet):
        with open(sheet, "r") as csvfile:
            reader = csv.DictReader(csvfile)
            first = next(reader)
            return list(set(key for key in first.keys()))

    def create_csv_with_fields_to_nuke(self, sheet, new_sheet):
        data = []
        with open(sheet, "r", newline="") as infile:
            reader = csv.DictReader(infile)
            for row in reader:
                row.update((key, "") for key in self.fields_to_nuke)
                data.append(row)
        with open(new_sheet, "w", newline="") as outfile:
            fieldnames = reader.fieldnames + self.fields_to_nuke
            writer = csv.DictWriter(outfile, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerows(data)
        return


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Read a csv and print the headers.")
    parser.add_argument(
        "-s", "--sheet", dest="sheet", help="Specify new csv.", required=True
    )
    parser.add_argument(
        "-o", "--old_sheet", dest="old_sheet", help="Specify old csv.", required=True
    )
    parser.add_argument(
        "-n", "--new_sheet", dest="new_sheet", help="Specify new csv.", required=True
    )
    args = parser.parse_args()
    ir = ImportRefactor(args.sheet, args.old_sheet)
    ir.create_csv_with_fields_to_nuke(args.sheet, args.new_sheet)



#!/usr/bin/env python3

from utk_exodus.finder import FileOrganizer
from utk_exodus.metadata import MetadataMapping
from utk_exodus.validate import ValidateMigration
from utk_exodus.controller import InterfaceController
from utk_exodus.template import ImportTemplate
from utk_exodus.combine import ImportRefactor
from utk_exodus.checksum import HashSheet
import click
import requests


@click.group()
def cli() -> None:
    pass


@cli.command("works", help="Create import sheet for works with metadata.")
@click.option(
    "--config",
    "-c",
    default="config/utk_dc.yml",
    help="Path to the configuration file for metadata mapping.",
)
@click.option(
    "--path",
    "-p",
    required=True,
    help="Path to the directory containing the metadata files.",
)
@click.option(
    "--output",
    "-o",
    default="delete/works.csv",
    help="Path to the output file you want to write to.",
)
def works(config: str, path: str, output: str) -> None:
    metadata = MetadataMapping(config, path)
    metadata.write_csv(output)
    r = requests.get(
        "https://raw.githubusercontent.com/utkdigitalinitiatives/m3_profiles/main/maps/utk.yml"
    )
    with open("delete/m3.yml", "wb") as f:
        f.write(r.content)
    print("Validating import ...")
    validator = ValidateMigration(profile="delete/m3.yml", migration_sheet=output)
    validator.iterate()


@cli.command(
    "add_files", help="Add filesets and attachments to and existing import sheet."
)
@click.option(
    "--sheet", "-s", help="Specify the original metadata works csv.", required=True
)
@click.option("--files_sheet", "-f", help="Optional: specify new sheet to write to.")
@click.option(
    "--what_to_add",
    "-w",
    help="Specify adding attachments, filesheets, or everything",
    default="everything",
)
@click.option(
    "--remote",
    "-r",
    help="Specify remote address of files",
    default="https://digital.lib.utk.edu/collections/islandora/object/",
)
def add_files(sheet: str, files_sheet: str, what_to_add: str, remote: str) -> None:
    if not files_sheet:
        files_sheet = f"{sheet.replace('.csv', '')}_with_filesets_and_attachments.csv"
    if what_to_add == "everything":
        what_to_add = ["filesets", "attachments"]
    """Take a CSV and Add files to it"""
    x = FileOrganizer(sheet, what_to_add, remote)
    x.write_csv(files_sheet)


@cli.command("works_and_files", help="Create import sheet with works and files.")
@click.option(
    "--config",
    "-c",
    default="utk_dc",
    type=click.Choice(
        ["utk_dc", "samvera", "utk_dc_no_uris", "utk_dc_no_uris_or_names"],
        case_sensitive=False,
    ),
    help="The configuration you want to use. By default, utk_dc.",
)
@click.option(
    "--collection",
    "-l",
    help="Specify the collection you want to download metadata for.",
)
@click.option(
    "--model",
    "-m",
    type=click.Choice(
        ["book", "image", "large_image", "pdf", "audio", "video"], case_sensitive=False
    ),
    help="The model you want to download metadata for.",
)
@click.option(
    "--path",
    "-p",
    help="Path to the directory containing the metadata files.",
)
@click.option(
    "--output",
    "-o",
    default="new_set",
    help="Path to the output directory you want to create.",
)
@click.option(
    "--remote",
    "-r",
    help="Specify remote address of files",
    default="https://digital.lib.utk.edu/collections/islandora/object/",
)
@click.option(
    "--total_size",
    "-t",
    help="Specify maximum number of attachments and filesets per sheet.",
    default=800,
)
def works_and_files(
    collection: str,
    config: str,
    model: str,
    path: str,
    output: str,
    remote: str,
    total_size: int,
) -> None:
    if model and collection:
        interface = InterfaceController(config, output, remote, total_size)
        interface.download_mods(collection, model)
    elif path:
        interface = InterfaceController(config, output, remote, total_size)
        interface.build_import_from_directory(path)
    else:
        print(
            "You must specify either a path to a directory or both a collection and model."
        )


@cli.command(
    "generate_sheet", help="Create full sample for content model from m3 profile."
)
@click.option(
    "--model",
    "-m",
    type=click.Choice(
        [
            "Attachment",
            "Audio",
            "Book",
            "CompoundObject",
            "GenericWork",
            "Image",
            "Newspaper",
            "PDF",
            "Video",
        ],
        case_sensitive=True,
    ),
    help="The content models to find rules about.",
)
@click.option(
    "--output",
    "-o",
    help="Specify where to write your output file.",
)
def generate_sheet(
    model: str,
    output: str,
) -> None:
    r = requests.get(
        "https://raw.githubusercontent.com/utkdigitalinitiatives/m3_profiles/main/maps/utk.yml"
    )
    with open("tmp/m3.yml", "wb") as f:
        f.write(r.content)
    it = ImportTemplate("tmp/m3.yml", model)
    it.write(output)


@cli.command(
    "remove_old_values", help="Remove old values from a previous import sheet."
)
@click.option(
    "--sheet",
    "-s",
    required=True,
    help="Specify the current sheet.",
)
@click.option(
    "--old_sheet",
    "-o",
    required=True,
    help="Specify the old sheet with unused values you want to remove.",
)
@click.option(
    "--new_sheet",
    "-n",
    required=True,
    help="Specify where you want to write the new sheet.",
)
def remove_old_values(
    sheet: str,
    old_sheet: str,
    new_sheet: str,
) -> None:
    ir = ImportRefactor(sheet, old_sheet)
    ir.create_csv_with_fields_to_nuke(sheet, new_sheet)
    print(f"Refactored sheet written to {new_sheet}.")


@cli.command(
    "hash_errors",
    help="Create sheet from a directory of errored import sheets.",
)
@click.option(
    "--path",
    "-p",
    required=True,
    help="Specify the path to the directory of sheets.",
)
@click.option(
    "--output",
    "-o",
    required=True,
    help="Specify where you want to write your sheets.",
)
def hash_errors(
    path: str,
    output: str,
) -> None:
    print(f"Generating checksums for bad files in csvs in {path}.")
    hs = HashSheet(path, output)
    hs.write()
    print(f"Hash sheet written to {output}.")



import requests


class FedoraObject:
    def __init__(self, auth, fedora_uri, pid):
        self.auth = auth
        self.fedora_uri = fedora_uri
        self.pid = f"{pid.replace('info:fedora/','').strip()}"

    @staticmethod
    def __guess_extension(content_type):
        mimetypes = {
            "image/tiff": "tif",
            "image/jp2": "jp2",
            "application/xml": "xml",
            "application/pdf": "pdf",
            "audio/vnd.wave": "wav",
            "image/jpeg": "jpg",
            "text/plain": "txt",
            "text/x-asm": "vtt",
            "text/x-c++": "vtt",
            "text/x-c": "vtt",
            "video/dv": "dv",
        }
        return mimetypes.get(content_type, "bin")

    def getDatastream(self, dsid, output):
        r = requests.get(
            f"{self.fedora_uri}/objects/{self.pid}/datastreams/{dsid}/content",
            auth=self.auth,
            allow_redirects=True,
        )
        if r.status_code == 200:
            open(
                f'{output}/{self.pid}_{dsid}.{self.__guess_extension(r.headers.get("Content-Type", "application/binary"))}',
                "wb",
            ).write(r.content)
        else:
            print(f"{r.status_code} on {self.pid}.")
        return



from .metadata import (
    DataProvider,
    ExtentProperty,
    GeoNamesProperty,
    KeywordProperty,
    LanguageURIProperty,
    LocalTypesProperties,
    MachineDate,
    MetadataMapping,
    NameProperty,
    PhysicalLocationsProperties,
    PublicationPlaceProperty,
    PublisherProperty,
    RightsOrLicenseProperties,
    RoleAndNameProperty,
    SubjectProperty,
    TitleProperty,
    TypesProperties,
)

__all__ = [
    "DataProvider",
    "ExtentProperty",
    "GeoNamesProperty",
    "KeywordProperty",
    "LanguageURIProperty",
    "LocalTypesProperties",
    "MachineDate",
    "MetadataMapping",
    "NameProperty",
    "PhysicalLocationsProperties",
    "PublicationPlaceProperty",
    "PublisherProperty",
    "RightsOrLicenseProperties",
    "RoleAndNameProperty",
    "SubjectProperty",
    "TitleProperty",
    "TypesProperties",
]



import click
import os
import requests
import shutil
from utk_exodus.finder import FileOrganizer
from utk_exodus.fedora import FedoraObject
from utk_exodus.curate import FileCurator
from utk_exodus.metadata import MetadataMapping
from utk_exodus.risearch import ResourceIndexSearch
from utk_exodus.restrict import RestrictionsSheet
from utk_exodus.validate import ValidateMigration
from pathlib import Path
from tqdm import tqdm


class InterfaceController:
    def __init__(self, config, output, remote, total_size):
        self.config = self.__load_config(config)
        self.output = output
        self.remote = remote
        self.total_size = total_size

    @staticmethod
    def __load_config(config):
        configs = {
            "utk_dc": "utk_dc.yml",
            "samvera": "samvera_default.yml",
            "utk_dc_no_uris": "utk_dc_no_uris.yml",
            "utk_dc_no_uris_or_names": "utk_no_uris_no_names.yml",
        }
        config_path = Path(__file__).parent / "../config"
        return config_path / configs.get(config)

    def __curate_filesets_and_attachments(self, csv_file):
        click.echo(
            click.style(
                "Curating filesets and attachments ...", fg="magenta", bold=True
            )
        )
        curator = FileCurator(f"{csv_file}")
        curator.write_files_and_attachments_only(
            f"{self.output}/{self.output.split('/')[-1]}.csv_filesheets_and_attachments_only.csv",
            attachments_per_sheet=int(self.total_size),
            multi_sheets="multi",
        )
        curator.write_works_and_collections_only(
            f"{self.output}/{self.output.split('/')[-1]}_works_and_collections_only.csv"
        )
        return

    def __generate_metadata_sheet(self, path):
        click.echo(click.style("Generating metadata sheet ...", fg="green", bold=True))
        os.makedirs(self.output, exist_ok=True)
        metadata = MetadataMapping(self.config, path)
        os.makedirs("tmp", exist_ok=True)
        metadata.write_csv("tmp/works.csv")
        return

    @staticmethod
    def __get_mods(collection, work_type):
        click.echo(click.style("Finding MODS files ...", fg="red", bold=True))
        risearch = ResourceIndexSearch().get_works_based_on_type_and_collection(
            work_type, collection
        )
        return risearch

    @staticmethod
    def __get_m3():
        r = requests.get(
            "https://raw.githubusercontent.com/utkdigitalinitiatives/m3_profiles/main/maps/utk.yml"
        )
        with open("tmp/m3.yml", "wb") as f:
            f.write(r.content)
        return

    def __grab_file_info(self):
        click.echo(click.style("Grabbing file info ...", fg="yellow", bold=True))
        x = FileOrganizer("tmp/works.csv", ["filesets", "attachments"], self.remote)
        x.write_csv(f"{self.output}/{self.output.split('/')[-1]}.csv")
        self.__get_m3()
        return

    def __validate_import(self):
        click.echo(click.style("Validating import ...", fg="blue", bold=True))
        validator = ValidateMigration(
            profile="tmp/m3.yml",
            migration_sheet=f"{self.output}/{self.output.split('/')[-1]}.csv",
        )
        validator.iterate()
        return

    def build_import_from_directory(self, path):
        self.__generate_metadata_sheet(path)
        self.__grab_file_info()
        self.__validate_import()
        self.__curate_filesets_and_attachments(
            f"{self.output}/{self.output.split('/')[-1]}.csv"
        )
        click.echo(click.style("Done ...", fg="cyan", bold=True))
        return

    @staticmethod
    def __get_policies(collection, work_type):
        click.echo(click.style("Finding Policy files ...", fg="red", bold=True))
        risearch = ResourceIndexSearch().get_policies_based_on_type_and_collection(
            work_type, collection
        )
        return risearch

    @staticmethod
    def __download(files, path, dsid):
        for record in tqdm(files):
            fedora = FedoraObject(
                auth=(
                    os.environ.get("FEDORA_USERNAME"),
                    os.environ.get("FEDORA_PASSWORD"),
                ),
                fedora_uri=os.environ.get("FEDORA_URI"),
                pid=f"{record.replace('info:fedora/', '').strip()}",
            )
            fedora.getDatastream(dsid=dsid, output=path)
        return

    def __download_policies(self, collection, work_type):
        policies = self.__get_policies(collection, work_type)
        os.makedirs("tmp/policy_downloads", exist_ok=True)
        os.makedirs("tmp/policy_downloads/current_collection", exist_ok=True)
        self.__download(policies, "tmp/policy_downloads/current_collection", "POLICY")
        return

    def __download_mods(self, collection, work_type):
        mods = self.__get_mods(collection, work_type)
        os.makedirs("tmp", exist_ok=True)
        os.makedirs("tmp/mods_downloads", exist_ok=True)
        os.makedirs("tmp/mods_downloads/current_collection", exist_ok=True)
        self.__download(mods, "tmp/mods_downloads/current_collection", "MODS")
        return

    def restrict_files_and_works(self, original_sheet, policies_location):
        click.echo(click.style("Restricting files and works ...", fg="red", bold=True))
        x = RestrictionsSheet(original_sheet, policies_location)
        x.write_csv(f"{self.output}/{self.output.split('/')[-1]}_visibility.csv")
        return

    def download_mods(self, collection, work_type):
        # @TODO: Rename this
        # @TODO: Make sure this works elsewhere
        self.__download_mods(collection, work_type)
        self.__generate_metadata_sheet("tmp/mods_downloads/current_collection")
        shutil.rmtree("tmp/mods_downloads/current_collection")
        self.__download_policies(collection, work_type)
        self.__grab_file_info()
        self.restrict_files_and_works(
            f"{self.output}/{self.output.split('/')[-1]}.csv",
            "tmp/policy_downloads/current_collection",
        )
        shutil.rmtree("tmp/policy_downloads/current_collection")
        self.__validate_import()
        self.__curate_filesets_and_attachments(
            f"{self.output}/{self.output.split('/')[-1]}_visibility.csv"
        )
        click.echo(click.style("Done ...", fg="cyan", bold=True))
        return



import csv


class FileCurator:
    def __init__(self, csv, curation_type="both"):
        self.original_csv = csv
        self.curation_type = curation_type
        self.files_and_attachments = self.__get_files_and_attachments_only()
        self.headers = self.__get_headers()

    def __get_headers(self):
        original_headers = [k for k, v in self.files_and_attachments[0].items()]
        return original_headers

    def __get_files_and_attachments_only(self):
        csv_content = []
        with open(self.original_csv, 'r') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                if self.curation_type == "both":
                    if row['model'] == "Attachment" or row['model'] == "FileSet":
                        csv_content.append(row)
                elif self.curation_type == "filesets":
                    if row['model'] == "FileSet":
                        csv_content.append(row)
                elif self.curation_type == "attachments":
                    if row['model'] == "Attachment":
                        csv_content.append(row)
        return csv_content

    def __get_works_and_collections(self):
        csv_content = []
        with open(self.original_csv, 'r') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                if row['model'] != "Attachment" and row['model'] != "FileSet":
                    csv_content.append(row)
        return csv_content


    def __write_sheet(self, filename, values, newline=''):
        with open(filename, 'w', newline=newline) as bulkrax_sheet:
            writer = csv.DictWriter(bulkrax_sheet, fieldnames=self.headers)
            writer.writeheader()
            for data in values:
                writer.writerow(data)
        return

    def write_files_and_attachments_only(self, base_filename, multi_sheets=False, attachments_per_sheet=500):
        if multi_sheets:
            bundles = [self.files_and_attachments[i:i + attachments_per_sheet] for i in range(0, len(self.files_and_attachments), attachments_per_sheet)]
            i = 0
            for bundle in bundles:
                self.__write_sheet(f"{base_filename.replace('.csv', '')}_{i}.csv", values=bundle)
                i += 1
            return
        else:
            self.__write_sheet(base_filename, values=self.files_and_attachments)
            return

    def write_works_and_collections_only(self, base_filename):
        self.__write_sheet(base_filename, values=self.__get_works_and_collections())
        return


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description='Add collections to sheet.')
    parser.add_argument("-s", "--sheet", dest="sheet", help="Specify the initial sheet.", required=True)
    parser.add_argument(
        "-f", "--files_sheet", dest="files_sheet", help="Optional: specify files sheet or files sheet pattern."
    )
    parser.add_argument(
        "-m", "--multi_sheets",
        dest="multi_sheets",
        help="multi or single", default="multi"
    )
    parser.add_argument(
        "-t", "--total_size",
        dest="total_size",
        help="If multisheet, the number of attachments and filesets. Must be even.", default=800
    )
    parser.add_argument(
        "-ty", "--type",
        dest="curation_type",
        help="filesets, attachments, both",
        default="both"
    )
    args = parser.parse_args()
    files_sheet = f"{args.sheet.replace('.csv', '')}_with_filesheets_and_attachments_only.csv"
    if args.files_sheet:
        files_sheet = args.files_sheet
    multi_sheet = True
    if args.multi_sheets == "single":
        multi_sheet = False
    x = FileCurator(args.sheet, curation_type=args.curation_type)
    x.write_files_and_attachments_only(
        files_sheet,
        multi_sheets=multi_sheet,
        attachments_per_sheet=int(args.total_size)
    )



from lxml import etree
import xmltodict


class BaseProperty:
    def __init__(self, path, namespaces):
        self.path = path
        self.namespaces = namespaces
        self.root = etree.parse(path)
        self.root_as_str = etree.tostring(self.root)


class StandardProperty(BaseProperty):
    def __init__(self, path, namespaces):
        super().__init__(path, namespaces)

    def find(self, xpaths):
        all_values = []
        for xpath in xpaths:
            matches = self.root.xpath(xpath, namespaces=self.namespaces)
            for match in matches:
                if not xpath.endswith("@xlink:href") and match.text is not None:
                    all_values.append(match.text)
                elif xpath.endswith("@xlink:href"):
                    all_values.append(match)
        return all_values


class XMLtoDictProperty:
    def __init__(self, file):
        self.path = file
        self.namespaces = {
            "http://www.loc.gov/mods/v3": "mods",
            "http://www.w3.org/1999/xlink": "xlink",
        }
        self.doc = self.__get_doc(file)

    def __get_doc(self, path):
        with open(path) as fd:
            return xmltodict.parse(
                fd.read(), process_namespaces=True, namespaces=self.namespaces
            )
